@@ -1,11 +1,9 @@
-/* Generate a file containing some preset patterns.
-   Print statistics for existing files.
+/* Multi-purpose tool for tar and cpio testsuite.
 
-   Copyright (C) 1995, 1996, 1997, 2001, 2003, 2004, 2005, 2006, 2007,
-   2008, 2009, 2016 Free Software Foundation, Inc.
+   Copyright (C) 1995-1997, 2001-2018 Free Software Foundation, Inc.
 
    François Pinard <pinard@iro.umontreal.ca>, 1995.
-   Sergey Poznyakoff <gray@mirddin.farlep.net>, 2004, 2005, 2006, 2007, 2008.
+   Sergey Poznyakoff <gray@gnu.org>, 2004-2018.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,8 +16,7 @@
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software Foundation,
-   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include <system.h>
@@ -43,6 +40,8 @@
 #ifndef EXIT_FAILURE
 # define EXIT_FAILURE 1
 #endif
+#define EXIT_USAGE 2
+#define EXIT_UNAVAILABLE 3
 
 #if ! defined SIGCHLD && defined SIGCLD
 # define SIGCHLD SIGCLD
@@ -96,10 +95,8 @@ size_t block_size = 512;
 /* Block buffer for sparse file */
 char *buffer;
 
-/* Number of arguments and argument vector for mode == mode_exec */
-int exec_argc;
-char **exec_argv;
-char *checkpoint_option;
+/* Checkpoint granularity for mode == mode_exec */
+char *checkpoint_granularity;
 
 /* Time for --touch option */
 struct timespec touch_time;
@@ -123,7 +120,7 @@ static char doc[] = N_("genfile manipulates data files for GNU paxutils test sui
 #define OPT_DATE       261
 #define OPT_VERBOSE    262
 #define OPT_SEEK       263
-#define OPT_UNLINK     264
+#define OPT_DELETE     264
 
 static struct argp_option options[] = {
 #define GRP 0
@@ -165,8 +162,8 @@ static struct argp_option options[] = {
   {NULL, 0, NULL, 0,
    N_("Synchronous execution options:"), GRP},
 
-  {"run", 'r', N_("OPTION"), OPTION_ARG_OPTIONAL,
-   N_("Execute ARGS. Useful with --checkpoint and one of --cut, --append, --touch, --unlink"),
+  {"run", 'r', N_("N"), OPTION_ARG_OPTIONAL,
+   N_("Execute ARGS. Trigger checkpoints every Nth record (default 1). Useful with --checkpoint and one of --cut, --append, --touch, --unlink"),
    GRP+1 },
   {"checkpoint", OPT_CHECKPOINT, N_("NUMBER"), 0,
    N_("Perform given action (see below) upon reaching checkpoint NUMBER"),
@@ -195,9 +192,10 @@ static struct argp_option options[] = {
   {"exec", OPT_EXEC, N_("COMMAND"), 0,
    N_("Execute COMMAND"),
    GRP+1 },
-  {"unlink", OPT_UNLINK, N_("FILE"), 0,
-   N_("Unlink FILE"),
+  {"delete", OPT_DELETE, N_("FILE"), 0,
+   N_("Delete FILE"),
    GRP+1 },
+  {"unlink", 0, 0, OPTION_ALIAS, NULL, GRP+1},
 #undef GRP
   { NULL, }
 };
@@ -246,15 +244,15 @@ get_size (const char *str, int allow_zero)
       if (9 < (unsigned) digit)
 	{
 	  if (xlat_suffix (&v, p))
-	    error (EXIT_FAILURE, 0, _("Invalid size: %s"), str);
+	    error (EXIT_USAGE, 0, _("Invalid size: %s"), str);
 	  else
 	    break;
 	}
       else if (x / 10 != v)
-	error (EXIT_FAILURE, 0, _("Number out of allowed range: %s"), str);
+	error (EXIT_USAGE, 0, _("Number out of allowed range: %s"), str);
       v = x + digit;
       if (v < 0)
-	error (EXIT_FAILURE, 0, _("Negative size: %s"), str);
+	error (EXIT_USAGE, 0, _("Negative size: %s"), str);
     }
   return v;
 }
@@ -274,7 +272,7 @@ verify_file (char *file_name)
 	       (unsigned long)st.st_size, (unsigned long)file_length);
 
       if (!quiet && mode == mode_sparse && !ST_IS_SPARSE (st))
-	error (EXIT_FAILURE, 0, _("created file is not sparse"));
+	error (EXIT_UNAVAILABLE, 0, _("created file is not sparse"));
     }
 }
 
@@ -289,7 +287,7 @@ struct action
   struct timespec ts;
 };
 
-static struct action *action_list;
+static struct action *action_head, *action_tail;
 
 void
 reg_action (int action, char *arg)
@@ -301,8 +299,12 @@ reg_action (int action, char *arg)
   act->ts = touch_time;
   act->size = file_length;
   act->name = arg;
-  act->next = action_list;
-  action_list = act;
+  act->next = NULL;
+  if (action_tail)
+    action_tail->next = act;
+  else
+    action_head = act;
+  action_tail = act;
 }
 
 static error_t
@@ -346,11 +348,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
 
     case 'r':
       mode = mode_exec;
-      if (arg)
-	{
-	  argcv_get (arg, "", NULL, &exec_argc, &exec_argv);
-	  checkpoint_option = "--checkpoint";
-	}
+      checkpoint_granularity = arg ? arg : "1";
       break;
 
     case 'T':
@@ -380,7 +378,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
     case OPT_TRUNCATE:
     case OPT_TOUCH:
     case OPT_EXEC:
-    case OPT_UNLINK:
+    case OPT_DELETE:
       reg_action (key, arg);
       break;
 
@@ -569,7 +567,7 @@ generate_sparse_file (int argc, char **argv)
   int flags = O_CREAT | O_RDWR | O_BINARY;
 
   if (!file_name)
-    error (EXIT_FAILURE, 0,
+    error (EXIT_USAGE, 0,
 	   _("cannot generate sparse files on standard output, use --file option"));
   if (!seek_offset)
     flags |= O_TRUNC;
@@ -661,13 +659,13 @@ print_stat (const char *name)
 	      if (*q)
 		{
 		  printf ("\n");
-		  error (EXIT_FAILURE, 0, _("incorrect mask (near `%s')"), q);
+		  error (EXIT_USAGE, 0, _("incorrect mask (near `%s')"), q);
 		}
 	    }
 	  else if (p[4])
 	    {
 	      printf ("\n");
-	      error (EXIT_FAILURE, 0, _("Unknown field `%s'"), p);
+	      error (EXIT_USAGE, 0, _("Unknown field `%s'"), p);
 	    }
 	  printf ("%0o", val);
 	}
@@ -700,7 +698,7 @@ print_stat (const char *name)
       else
 	{
 	  printf ("\n");
-	  error (EXIT_FAILURE, 0, _("Unknown field `%s'"), p);
+	  error (EXIT_USAGE, 0, _("Unknown field `%s'"), p);
 	}
       p = strtok (NULL, ",");
       if (p)
@@ -769,9 +767,19 @@ exec_checkpoint (struct action *p)
 	error (0, 0, _("command failed: %s"), p->name);
       break;
 
-    case OPT_UNLINK:
-      if (unlink (p->name))
-	error (0, errno, _("cannot unlink `%s'"), p->name);
+    case OPT_DELETE:
+      {
+	struct stat st;
+	if (stat (p->name, &st))
+	  error (0, errno, _("cannot stat `%s'"), p->name);
+	else if (S_ISDIR (st.st_mode))
+	  {
+	    if (rmdir (p->name))
+	      error (0, errno, _("cannot remove directory `%s'"), p->name);
+	  }
+	else if (unlink (p->name))
+	  error (0, errno, _("cannot unlink `%s'"), p->name);
+      }
       break;
 
     default:
@@ -784,7 +792,7 @@ process_checkpoint (size_t n)
 {
   struct action *p, *prev = NULL;
 
-  for (p = action_list; p; )
+  for (p = action_head; p; )
     {
       struct action *next = p->next;
 
@@ -795,7 +803,9 @@ process_checkpoint (size_t n)
 	  if (prev)
 	    prev->next = next;
 	  else
-	    action_list = next;
+	    action_head = next;
+	  if (next == NULL)
+	    action_tail = prev;
 	  free (p);
 	}
       else
@@ -805,10 +815,10 @@ process_checkpoint (size_t n)
     }
 }
 
-#define CHECKPOINT_TEXT "Write checkpoint"
+#define CHECKPOINT_TEXT "genfile checkpoint"
 
 void
-exec_command (void)
+exec_command (int argc, char **argv)
 {
   int status;
   pid_t pid;
@@ -816,20 +826,28 @@ exec_command (void)
   char *p;
   FILE *fp;
   char buf[128];
-
+  int xargc;
+  char **xargv;
+  int i;
+  char checkpoint_option[80];
+  
   /* Insert --checkpoint option.
-     FIXME: This assumes that exec_argv does not use traditional tar options
+     FIXME: This assumes that argv does not use traditional tar options
      (without dash).
-     FIXME: There is no way to set checkpoint argument (granularity).
   */
-  if (checkpoint_option)
-    {
-      exec_argc++;
-      exec_argv = xrealloc (exec_argv, (exec_argc + 1) * sizeof (*exec_argv));
-      memmove (exec_argv+2, exec_argv+1,
-	       (exec_argc - 1) * sizeof (*exec_argv));
-      exec_argv[1] = checkpoint_option;
-    }
+  xargc = argc + 5;
+  xargv = xcalloc (xargc + 1, sizeof (xargv[0]));
+  xargv[0] = argv[0];
+  snprintf (checkpoint_option, sizeof (checkpoint_option),
+	    "--checkpoint=%s", checkpoint_granularity);
+  xargv[1] = checkpoint_option;
+  xargv[2] = "--checkpoint-action";
+  xargv[3] = "echo=" CHECKPOINT_TEXT " %u";
+  xargv[4] = "--checkpoint-action";
+  xargv[5] = "wait=SIGUSR1";
+
+  for (i = 1; i <= argc; i++)
+    xargv[i + 5] = argv[i];
 
 #ifdef SIGCHLD
   /* System V fork+wait does not work if SIGCHLD is ignored.  */
@@ -855,8 +873,8 @@ exec_command (void)
       /* Make sure POSIX locale is used */
       setenv ("LC_ALL", "POSIX", 1);
 
-      execvp (exec_argv[0], exec_argv);
-      error (EXIT_FAILURE, errno, "execvp %s", exec_argv[0]);
+      execvp (xargv[0], xargv);
+      error (EXIT_FAILURE, errno, "execvp %s", xargv[0]);
     }
 
   /* Master */
@@ -883,6 +901,7 @@ exec_command (void)
 	      if (!(*end && !isspace ((unsigned char) *end)))
 		{
 		  process_checkpoint (n);
+		  kill (pid, SIGUSR1);
 		  continue;
 		}
 	    }
@@ -935,7 +954,7 @@ main (int argc, char **argv)
   /* Decode command options.  */
 
   if (argp_parse (&argp, argc, argv, 0, &index, NULL))
-    exit (EXIT_FAILURE);
+    exit (EXIT_USAGE);
 
   argc -= index;
   argv += index;
@@ -944,7 +963,7 @@ main (int argc, char **argv)
     {
     case mode_stat:
       if (argc == 0)
-	error (EXIT_FAILURE, 0, _("--stat requires file names"));
+	error (EXIT_USAGE, 0, _("--stat requires file names"));
 
       while (argc--)
 	print_stat (*argv++);
@@ -957,7 +976,7 @@ main (int argc, char **argv)
 
     case mode_generate:
       if (argc)
-	error (EXIT_FAILURE, 0, _("too many arguments"));
+	error (EXIT_USAGE, 0, _("too many arguments"));
       if (files_from)
 	generate_files_from_list ();
       else
@@ -968,14 +987,7 @@ main (int argc, char **argv)
       break;
 
     case mode_exec:
-      if (!checkpoint_option)
-	{
-	  exec_argc = argc;
-	  exec_argv = argv;
-	}
-      else if (argc)
-	error (EXIT_FAILURE, 0, _("too many arguments"));
-      exec_command ();
+      exec_command (argc, argv);
       break;
 
     default:
